__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from ..frameworks import BaseTorchEncoder
from ...decorators import batching, as_ndarray


class FarmTextEncoder(BaseTorchEncoder):
    """FARM-based text encoder: (Framework for Adapting Representation Models)
    https://github.com/deepset-ai/FARM

    It encodes an array of string in size `B` into an ndarray in size `B x D`
    """

    def __init__(self, model_name: str = 'deepset/bert-base-cased-squad2',
                 num_processes: int = 0, extraction_strategy: str = 'cls_token',
                 extraction_layer: int = -1,
                 *args,
                 **kwargs):
        """

        :param model_name_or_path:  Local directory or public name of the model to load.
        :param num_processes: the number of processes for `multiprocessing.Pool`. Set to value of 0 to disable
                              multiprocessing. Set to None to let Inferencer use all CPU cores. If you want to
                              debug the Language Model, you might need to disable multiprocessing!
        :param extraction_strategy: Strategy to extract vectors. Choices: 'cls_token' (sentence vector), 'reduce_mean'
                               (sentence vector), reduce_max (sentence vector), 'per_token' (individual token vectors)
        :param extraction_layer: number of layer from which the embeddings shall be extracted. Default: -1 (very last layer).
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.model_name = 'deepset/bert-base-cased-squad2' or model_name
        self.num_processes = num_processes
        self.extraction_strategy = extraction_strategy
        self.extraction_layer = extraction_layer

    def post_init(self):
        from farm.infer import Inferencer
        self.model = Inferencer.load(model_name_or_path=self.model_name, task_type='embeddings',
                                     num_processes=self.num_processes)

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        basic_texts = [{'text': s} for s in data]
        embeds = np.stack([k['vec'] for k in self.model.extract_vectors(dicts=basic_texts)])
        return embeds



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterator, Optional, Any

from .. import BaseRecursiveDriver, BaseDriver
from ...proto import jina_pb2


class QueryLangDriver(BaseRecursiveDriver):
    """
    :class:`QueryLangDriver` allows a driver to read arguments from the protobuf message. This allows a
    driver to override its behavior based on the message it receives. Extremely useful in production, for example,
    get ``top_k`` results, doing pagination, filtering.

    To register the field you want to read from the message, simply register them in :meth:`__init__`.
    For example, ``__init__(self, arg1, arg2, **kwargs)`` will allow the driver to read field ``arg1`` and ``arg2`` from
    the message. When they are not found in the message, the value ``_arg1`` and ``_arg2`` will be used. Note the underscore
    prefix.

    .. note::
        - To set default value of ``arg1``, use ``self._arg1 = ``, note the underscore in the front.
        - To access ``arg1``, simply use ``self.arg1``. It automatically switch between default ``_arg1`` and ``arg1`` from the request.

    For successful value reading, the following condition must be met:

        - the ``name`` in the proto must match with the current class name
        - the ``disabled`` field in the proto should not be ``False``
        - the ``priority`` in the proto should be strictly greater than the driver's priority (by default is 0)
        - the field name must exist in proto's ``parameters``

    """

    def __init__(self, priority: int = 0, *args, **kwargs):
        """

        :param priority: the priority of its default arg values (hardcoded in Python). If the
        received ``QueryLang`` has a higher priority, it will override the hardcoded value
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self._priority = priority

    def _get_parameter(self, key: str, default: Any):
        for q in self.queryset:
            if (not q.disabled and self.__class__.__name__ == q.name and
                    q.priority > self._priority and key in q.parameters):
                return q.parameters[key]
        return getattr(self, f'_{key}', default)

    def __getattr__(self, name: str):
        # https://docs.python.org/3/reference/datamodel.html#object.__getattr__
        if name == '_init_kwargs_dict':
            # raise attribute error to avoid recursive call
            raise AttributeError
        if name in self._init_kwargs_dict:
            return self._get_parameter(name, default=self._init_kwargs_dict[name])


class BaseQueryLangDriver(BaseDriver):
    """
    :class:`BaseQueryLangDriver` adds native data querying capabilities to Jina,

    The following Standard Query Operator to be implemented
    - filter/where: filter the doc/chunk by its attributes
    - select/exclude: select attributes to include in the results
    - limit/take/slicing: take the first k doc/chunk
    - sort/order_by: sort the doc/chunk
    - reverse: reverse the list of collections
    """

    def __init__(self, raw_query=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if raw_query:
            self.driver_query = self.parse(raw_query)

    def __call__(self, *args, **kwargs):
        f = getattr(self.req, 'raw_query', None)
        req_query = self.parse(f) if f else None
        filtered = self.execute(self.req.docs, self.driver_query, req_query)
        if filtered is None:
            # filter is done in-place.
            pass
        elif isinstance(filtered, Iterator):
            # count length before
            len_before = len(self.req.docs)
            for d in filtered:
                _d = self.req.docs.add()
                _d.CopyFrom(d)
            # remove old ones
            del self.req.docs[:len_before]
        else:
            raise ValueError(f'do not support {type(filtered)} return type')

    def parse(self, raw_filter):
        raise NotImplementedError

    def execute(self, docs: Iterator['jina_pb2.Document'], driver_query, req_query) -> Optional[
        Iterator['jina_pb2.Document']]:
        """

        :param docs: the document to be filtered
        :param driver_query: the parsed filter assigned to the driver
        :param req_query: the parsed filter carried on the query
        :return: either an iterator of document,
        """
        raise NotImplementedError


class GraphQLDriver(BaseQueryLangDriver):
    """
    GraphQL as in-memory filter
    """


class CythonFilterDriver(BaseQueryLangDriver):
    """
    Cython implementation of the in-memory proto filter
    """



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable

from . import QueryLangDriver

if False:
    from ...proto import jina_pb2


class ReverseQL(QueryLangDriver):
    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        prev_len = len(docs)
        for d in reversed(docs):
            dd = docs.add()
            dd.CopyFrom(d)
        del docs[:prev_len]



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

from . import QueryLangDriver

if False:
    from ...proto import jina_pb2


class ExcludeQL(QueryLangDriver):
    """Clean some fields from the chunk-level protobuf to reduce the total size of the request
    """

    def __init__(self, fields: Tuple, *args, **kwargs):
        """

        :param fields: the pruned field names in tuple
        """
        super().__init__(*args, **kwargs)
        if isinstance(fields, str):
            self._fields = {fields, }
        else:
            self._fields = set(fields)

        # for deleting field in a recursive structure, postorder is safer
        self.recursion_order = 'post'

    def _apply(self, doc: 'jina_pb2.Document', *args, **kwargs):
        for k in self.fields:
            doc.ClearField(k)


class SelectQL(ExcludeQL):
    def _apply(self, doc: 'jina_pb2.Document', *args, **kwargs):
        for k in doc.DESCRIPTOR.fields_by_name.keys():
            if k not in self.fields:
                doc.ClearField(k)


# ChunkPruneDriver: pruned=('embedding', 'buffer', 'blob', 'text')
# DocPruneDriver: pruned=('chunks', 'buffer')

class ExcludeReqQL(ExcludeQL):
    """Clean up request from the protobuf message to reduce the total size of the message

        This is often useful when the proceeding Pods require only a signal, not the full message.
    """

    def __call__(self, *args, **kwargs):
        for k in self.fields:
            self.msg.ClearField(k)


class SelectReqQL(ExcludeReqQL):
    def __call__(self, *args, **kwargs):
        for k in self.msg.DESCRIPTOR.fields_by_name.keys():
            if k not in self.fields:
                self.msg.ClearField(k)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable

from . import QueryLangDriver

if False:
    from ...proto import jina_pb2


class SliceQL(QueryLangDriver):
    """Restrict the size of the ``matches`` to ``k`` (given by the request)

    This driver works on both chunk and doc level
    """

    def __init__(self, start: int, end: int = None, *args, **kwargs):
        """

        :param start: Zero-based index at which to start extraction.
        :param end:  Zero-based index before which to end extraction.
                slice extracts up to but not including end. For example, take(1,4) extracts
                the second element through the fourth element (elements indexed 1, 2, and 3).
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self._start = int(start)
        self._end = int(end)

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        if self.start <= 0 and (self.end is None or self.end >= len(docs)):
            pass
        else:
            del docs[int(self.end):]
            del docs[:int(self.start)]



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, Any, Iterable

from .queryset.lookup import Q
from . import QueryLangDriver

if False:
    from ...proto import jina_pb2


class FilterQL(QueryLangDriver):
    def __init__(self, lookups: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lookups = Q(**lookups) if lookups else None

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        if self.lookups:
            miss_idx = []
            for idx, doc in enumerate(docs):
                if not self.lookups.evaluate(doc):
                    miss_idx.append(idx)

            # delete non-exit matches in reverse
            for j in reversed(miss_idx):
                del docs[j]



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable

from . import QueryLangDriver
from ...helper import rgetattr

if False:
    from ...proto import jina_pb2


class SortQL(QueryLangDriver):
    """Restrict the size of the ``matches`` to ``k`` (given by the request)

    This driver works on both chunk and doc level
    """

    def __init__(self, field: str, reverse: bool = False, *args, **kwargs):
        """

        :param reverse: sort the value from big to small
        """

        super().__init__(*args, **kwargs)
        self._reverse = reverse
        self._field = field

    def _apply_all(self, docs: Iterable['jina_pb2.Document'], *args, **kwargs):
        docs.sort(key=lambda x: rgetattr(x, self.field), reverse=self.reverse)

